﻿using System;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;

namespace Kaleidovideo
{
    public class KaleidoscopeVideoSink : VideoSink
    {
        WriteableBitmap bitmap;
        VideoFormat videoFormat;
        int[] indexMap;             // maps bitmap pixel index to video pixel index

        public KaleidoscopeVideoSink(WriteableBitmap bitmap)
        {
            this.bitmap = bitmap;
            indexMap = new int[bitmap.PixelWidth * bitmap.PixelHeight];
        }

        protected override void OnCaptureStarted() { }

        protected override void OnCaptureStopped() { }

        protected override void OnFormatChange(VideoFormat videoFormat)
        {
            this.videoFormat = videoFormat;

            // Define a source equilateral triangle of the video 
            int side = (int)(videoFormat.PixelHeight / Math.Sqrt(0.75));
            int xCenter = videoFormat.PixelWidth / 2;
            Point A = new Point(xCenter - side / 2, 0);                     // top left vertex
            Point B = new Point(xCenter + side / 2, 0);                     // top right vertex
            Point C = new Point(xCenter, videoFormat.PixelHeight - 1);      // bottom vertex

            // Go down the row and call GenerateIndexMap for each triangle
            GenerateIndexMap(0, 0, 104, A, B, C);
            GenerateIndexMap(104, 60, 0, C, A, B);
            GenerateIndexMap(0, 120, 104, B, C, A);
            GenerateIndexMap(104, 180, 0, A, B, C);
            GenerateIndexMap(0, 240, 104, C, A, B);
            GenerateIndexMap(104, 300, 0, B, C, A);
            GenerateIndexMap(0, 360, 104, A, B, C);
            GenerateIndexMap(104, 420, 0, C, A, B);
            GenerateIndexMap(0, 480, 104, B, C, A);
        }

        // First three arguments reference the WritableBitmap destination;
        //  Second three arguments reference a triangle in the video source
        void GenerateIndexMap(int xApex, int yApex, int xBase, Point ptApex, Point ptTopBase, Point ptBotBase)
        {
            // Derive information about source triangle
            Point ptBaseCenter = new Point((ptTopBase.X + ptBotBase.X) / 2,
                                           (ptTopBase.Y + ptBotBase.Y) / 2);
            Point vector = new Point(ptBotBase.X - ptTopBase.X,
                                     ptBotBase.Y - ptTopBase.Y);

            // From apex of destination triangle to base
            for (int x = 0; x <= 104; x++)
            {
                int xDst = ((104 - x) * xApex + x * xBase) / 104;
                int y1 = -60 * x / 104;
                int y2 = 60 * x / 104;
                double xRel = x / 104.0;

                // Across the triangle
                for (int y = y1; y <= y2; y++)
                {
                    int yDst = yApex + y;
                    double yRel = y / 120.0;

                    if (yDst >= 0 && yDst < bitmap.PixelHeight)
                    {
                        Point ptAltitude = new Point((1 - xRel) * ptApex.X + xRel * ptBaseCenter.X,
                                                     (1 - xRel) * ptApex.Y + xRel * ptBaseCenter.Y);
                        Point ptSource = new Point(ptAltitude.X + yRel * vector.X, 
                                                   ptAltitude.Y + yRel * vector.Y);
                        int xSrc = (int)ptSource.X;
                        int ySrc = (int)ptSource.Y;
                        int indexSrc = ySrc * videoFormat.PixelWidth + xSrc;
                        int indexDst = bitmap.PixelWidth * yDst + xDst;
                        indexMap[indexDst] = indexSrc;
                    }
                }
            }
        }

        protected override void OnSample(long sampleTimeInHundredNanoseconds, long frameDurationInHundredNanoseconds, byte[] sampleData)
        {
            for (int bitmapIndex = 0; bitmapIndex < indexMap.Length; bitmapIndex++)
            {
                // Assumes 32-bit color
                int videoIndex = 4 * indexMap[bitmapIndex];

                int pixel = (sampleData[videoIndex + 0] << 0) |
                            (sampleData[videoIndex + 1] << 8) |
                            (sampleData[videoIndex + 2] << 16) |
                            (sampleData[videoIndex + 3] << 24);

                bitmap.Pixels[bitmapIndex] = pixel;
            }

            bitmap.Dispatcher.BeginInvoke(() => { bitmap.Invalidate(); });
        }

    }
}
